"""Orchestrator for managing pactown service ecosystems."""

import time
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import httpx
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from .config import EcosystemConfig, load_config
from .network import ServiceRegistry
from .resolver import DependencyResolver
from .sandbox_manager import SandboxManager, ServiceProcess

console = Console()


@dataclass
class ServiceHealth:
    """Health status of a service."""
    name: str
    healthy: bool
    status_code: Optional[int] = None
    response_time_ms: Optional[float] = None
    error: Optional[str] = None


class Orchestrator:
    """Orchestrates the lifecycle of a pactown ecosystem."""

    def __init__(
        self,
        config: EcosystemConfig,
        base_path: Optional[Path] = None,
        verbose: bool = True,
        dynamic_ports: bool = True,
    ):
        self.config = config
        self.base_path = base_path or Path.cwd()
        self.verbose = verbose
        self.dynamic_ports = dynamic_ports
        self.resolver = DependencyResolver(config)
        self.sandbox_manager = SandboxManager(config.sandbox_root)
        self._running: dict[str, ServiceProcess] = {}

        # Service registry for dynamic port allocation and discovery
        registry_path = Path(config.sandbox_root) / ".pactown-services.json"
        self.service_registry = ServiceRegistry(storage_path=registry_path)

    @classmethod
    def from_file(
        cls,
        config_path: str | Path,
        verbose: bool = True,
        dynamic_ports: bool = True,
    ) -> "Orchestrator":
        """Create orchestrator from configuration file."""
        config_path = Path(config_path)
        config = load_config(config_path)
        return cls(config, base_path=config_path.parent, verbose=verbose, dynamic_ports=dynamic_ports)

    def _get_readme_path(self, service_name: str) -> Path:
        """Get the README path for a service."""
        service = self.config.services[service_name]
        readme_path = self.base_path / service.readme
        if not readme_path.exists():
            raise FileNotFoundError(f"README not found: {readme_path}")
        return readme_path

    def validate(self) -> bool:
        """Validate the ecosystem configuration."""
        issues = self.resolver.validate()

        for name, service in self.config.services.items():
            readme_path = self.base_path / service.readme
            if not readme_path.exists():
                issues.append(f"README not found for '{name}': {readme_path}")

        if issues:
            console.print("[red]Validation failed:[/red]")
            for issue in issues:
                console.print(f"  • {issue}")
            return False

        console.print("[green]✓ Ecosystem configuration is valid[/green]")
        return True

    def start_service(self, service_name: str) -> ServiceProcess:
        """Start a single service."""
        if service_name not in self.config.services:
            raise ValueError(f"Unknown service: {service_name}")

        service = self.config.services[service_name]
        readme_path = self._get_readme_path(service_name)

        # Register service and get allocated port
        if self.dynamic_ports:
            endpoint = self.service_registry.register(
                name=service_name,
                preferred_port=service.port,
                health_check=service.health_check,
            )
            actual_port = endpoint.port

            if self.verbose and actual_port != service.port:
                console.print(f"  [yellow]Port {service.port} busy, using {actual_port}[/yellow]")
        else:
            actual_port = service.port

        # Get dependencies from config
        dep_names = [d.name for d in service.depends_on]

        # Build environment with service discovery
        env = self.service_registry.get_environment(service_name, dep_names)

        # Add any extra env from config
        env.update(service.env)
        env["PACTOWN_ECOSYSTEM"] = self.config.name

        # Override port in service config for this run
        service_copy = service
        if actual_port != service.port:
            from dataclasses import replace
            service_copy = replace(service, port=actual_port)

        process = self.sandbox_manager.start_service(
            service_copy, readme_path, env, verbose=self.verbose
        )
        self._running[service_name] = process
        return process

    def start_all(
        self,
        wait_for_health: bool = True,
        parallel: bool = True,
        max_workers: int = 4,
    ) -> dict[str, ServiceProcess]:
        """
        Start all services in dependency order.

        Args:
            wait_for_health: Wait for health checks
            parallel: Use parallel execution for independent services
            max_workers: Max parallel workers
        """
        if parallel:
            return self._start_all_parallel(wait_for_health, max_workers)
        else:
            return self._start_all_sequential(wait_for_health)

    def _start_all_sequential(self, wait_for_health: bool = True) -> dict[str, ServiceProcess]:
        """Start all services sequentially in dependency order."""
        order = self.resolver.get_startup_order()

        if self.verbose:
            console.print(f"\n[bold]Starting ecosystem: {self.config.name}[/bold]")
            console.print(f"Startup order: {' → '.join(order)}\n")

        for name in order:
            try:
                self.start_service(name)

                if wait_for_health:
                    service = self.config.services[name]
                    if service.health_check:
                        self._wait_for_health(name, timeout=service.timeout)
                    else:
                        time.sleep(1)

            except Exception as e:
                console.print(f"[red]Failed to start {name}: {e}[/red]")
                self.stop_all()
                raise

        if self.verbose:
            self.print_status()

        return self._running

    def _start_all_parallel(
        self,
        wait_for_health: bool = True,
        max_workers: int = 4,
    ) -> dict[str, ServiceProcess]:
        """
        Start services in parallel waves based on dependencies.

        Services with no unmet dependencies start together in parallel.
        Once a wave completes, the next wave starts.
        """
        from concurrent.futures import ThreadPoolExecutor, as_completed

        if self.verbose:
            console.print(f"\n[bold]Starting ecosystem: {self.config.name} (parallel)[/bold]")

        # Build dependency map
        deps_map: dict[str, list[str]] = {}
        for name, service in self.config.services.items():
            deps_map[name] = [d.name for d in service.depends_on if d.name in self.config.services]

        started = set()
        remaining = set(self.config.services.keys())
        wave_num = 0

        while remaining:
            # Find services ready to start (all deps satisfied)
            ready = [
                name for name in remaining
                if all(d in started for d in deps_map.get(name, []))
            ]

            if not ready:
                raise ValueError(f"Cannot resolve dependencies for: {remaining}")

            wave_num += 1
            if self.verbose:
                console.print(f"\n[cyan]Wave {wave_num}:[/cyan] {', '.join(ready)}")

            # Start ready services in parallel
            wave_results = {}
            wave_errors = {}

            with ThreadPoolExecutor(max_workers=min(max_workers, len(ready))) as executor:
                futures = {}

                for name in ready:
                    future = executor.submit(self._start_service_with_health, name, wait_for_health)
                    futures[future] = name

                for future in as_completed(futures):
                    name = futures[future]
                    try:
                        proc = future.result()
                        wave_results[name] = proc
                        self._running[name] = proc
                        started.add(name)
                        remaining.remove(name)
                    except Exception as e:
                        wave_errors[name] = str(e)
                        remaining.remove(name)

            # Report wave results
            for name in wave_results:
                if self.verbose:
                    console.print(f"  [green]✓[/green] {name} started")

            for name, error in wave_errors.items():
                console.print(f"  [red]✗[/red] {name}: {error}")

            # Stop on any failure
            if wave_errors:
                console.print("\n[red]Stopping due to errors...[/red]")
                self.stop_all()
                raise RuntimeError(f"Failed to start services: {wave_errors}")

        if self.verbose:
            console.print()
            self.print_status()

        return self._running

    def _start_service_with_health(self, service_name: str, wait_for_health: bool) -> ServiceProcess:
        """Start a service and optionally wait for health check."""
        proc = self.start_service(service_name)

        if wait_for_health:
            service = self.config.services[service_name]
            if service.health_check:
                self._wait_for_health(service_name, timeout=service.timeout)
            else:
                time.sleep(0.5)

        return proc

    def stop_service(self, service_name: str) -> bool:
        """Stop a single service."""
        if service_name not in self._running:
            return False

        if self.verbose:
            console.print(f"Stopping {service_name}...")

        success = self.sandbox_manager.stop_service(service_name)
        if success:
            del self._running[service_name]
        return success

    def stop_all(self) -> None:
        """Stop all services in reverse dependency order."""
        order = self.resolver.get_shutdown_order()

        if self.verbose:
            console.print(f"\n[bold]Stopping ecosystem: {self.config.name}[/bold]")

        for name in order:
            if name in self._running:
                self.stop_service(name)
            # Unregister from service registry
            self.service_registry.unregister(name)

        self.sandbox_manager.stop_all()
        self._running.clear()

    def restart_service(self, service_name: str) -> ServiceProcess:
        """Restart a single service."""
        self.stop_service(service_name)
        time.sleep(0.5)
        return self.start_service(service_name)

    def check_health(self, service_name: str) -> ServiceHealth:
        """Check health of a service."""
        if service_name not in self.config.services:
            return ServiceHealth(name=service_name, healthy=False, error="Unknown service")

        service = self.config.services[service_name]

        if service_name not in self._running:
            return ServiceHealth(name=service_name, healthy=False, error="Not running")

        if not self._running[service_name].is_running:
            return ServiceHealth(name=service_name, healthy=False, error="Process died")

        if not service.health_check:
            return ServiceHealth(name=service_name, healthy=True)

        # Get actual port from registry (may differ from config if dynamic)
        endpoint = self.service_registry.get(service_name)
        port = endpoint.port if endpoint else service.port
        url = f"http://localhost:{port}{service.health_check}"

        try:
            start = time.time()
            response = httpx.get(url, timeout=5.0)
            elapsed = (time.time() - start) * 1000

            return ServiceHealth(
                name=service_name,
                healthy=response.status_code < 400,
                status_code=response.status_code,
                response_time_ms=elapsed,
            )
        except Exception as e:
            return ServiceHealth(
                name=service_name,
                healthy=False,
                error=str(e),
            )

    def _wait_for_health(self, service_name: str, timeout: int = 60) -> bool:
        """Wait for a service to become healthy."""
        service = self.config.services[service_name]

        if not service.health_check:
            return True

        # Get actual port from registry
        endpoint = self.service_registry.get(service_name)
        port = endpoint.port if endpoint else service.port
        url = f"http://localhost:{port}{service.health_check}"
        deadline = time.time() + timeout

        while time.time() < deadline:
            try:
                response = httpx.get(url, timeout=2.0)
                if response.status_code < 400:
                    if self.verbose:
                        console.print(f"  [green]✓[/green] {service_name} is healthy")
                    return True
            except Exception:
                pass
            time.sleep(0.5)

        console.print(f"  [yellow]⚠[/yellow] {service_name} health check timed out")
        return False

    def print_status(self) -> None:
        """Print status of all services."""
        table = Table(title=f"Ecosystem: {self.config.name}")
        table.add_column("Service", style="cyan")
        table.add_column("Port", style="blue")
        table.add_column("Status", style="green")
        table.add_column("PID")
        table.add_column("Health")

        for name, service in self.config.services.items():
            # Get actual port from registry
            endpoint = self.service_registry.get(name)
            actual_port = endpoint.port if endpoint else service.port

            if name in self._running:
                proc = self._running[name]
                running = "🟢 Running" if proc.is_running else "🔴 Stopped"
                pid = str(proc.pid)

                health = self.check_health(name)
                if health.healthy:
                    health_str = f"✓ {health.response_time_ms:.0f}ms" if health.response_time_ms else "✓"
                else:
                    health_str = health.error or "✗"
            else:
                running = "⚪ Not started"
                pid = "-"
                health_str = "-"
                actual_port = service.port  # Use config port if not registered

            table.add_row(
                name,
                str(actual_port) if actual_port else "-",
                running,
                pid,
                health_str,
            )

        console.print(table)

    def print_graph(self) -> None:
        """Print dependency graph."""
        console.print(Panel(self.resolver.print_graph(), title="Dependency Graph"))

    def get_logs(self, service_name: str, lines: int = 100) -> Optional[str]:
        """Get recent logs from a service (if available)."""
        if service_name not in self._running:
            return None

        proc = self._running[service_name]
        if proc.process and proc.process.stdout:
            return proc.process.stdout.read()
        return None


def run_ecosystem(config_path: str | Path, wait: bool = True) -> Orchestrator:
    """Convenience function to start an ecosystem."""
    orch = Orchestrator.from_file(config_path)

    if not orch.validate():
        raise ValueError("Invalid ecosystem configuration")

    orch.start_all()

    if wait:
        try:
            console.print("\n[dim]Press Ctrl+C to stop all services[/dim]\n")
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            console.print("\n[yellow]Shutting down...[/yellow]")
            orch.stop_all()

    return orch



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest

from pactown.config import ServiceConfig
from pactown.sandbox_manager import SandboxManager


def test_node_project_uses_npm_instead_of_pip_even_if_deps_lang_is_wrong(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    sandbox_root = tmp_path / "sandboxes"
    manager = SandboxManager(sandbox_root)

    monkeypatch.setattr(manager._dep_cache, "get_cached_venv", lambda _deps: None)

    import pactown.sandbox_manager as sm_module

    captured: dict[str, object] = {"npm": 0, "pip": 0}

    def fake_run(cmd, capture_output=False, text=False, check=False, env=None, cwd=None, **kwargs):
        cmd0 = str(cmd[0]) if isinstance(cmd, list) and cmd else str(cmd)
        if isinstance(cmd, list) and cmd0 == "npm":
            captured["npm"] = int(captured["npm"]) + 1
            return SimpleNamespace(returncode=0, stdout="", stderr="")
        if isinstance(cmd, list) and "pip" in cmd0 and "install" in cmd:
            captured["pip"] = int(captured["pip"]) + 1
            return SimpleNamespace(returncode=0, stdout="", stderr="")
        return SimpleNamespace(returncode=0, stdout="", stderr="")

    monkeypatch.setattr(sm_module.subprocess, "run", fake_run)

    readme = """# Express Hello

```javascript markpact:file path=index.js
console.log('hello');
```

```python markpact:deps
express
```

```bash markpact:run
node index.js $PORT
```
"""

    readme_path = tmp_path / "README.md"
    readme_path.write_text(readme)

    service = ServiceConfig(name="svc", readme=str(readme_path), port=8000)

    sandbox = manager.create_sandbox(
        service=service,
        readme_path=readme_path,
        install_dependencies=True,
        on_log=None,
        env=None,
    )

    assert (Path(sandbox.path) / "package.json").exists()
    assert int(captured["npm"]) == 1
    assert int(captured["pip"]) == 0


def test_node_deps_block_creates_package_json_and_calls_npm(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    sandbox_root = tmp_path / "sandboxes"
    manager = SandboxManager(sandbox_root)

    monkeypatch.setattr(manager._dep_cache, "get_cached_venv", lambda _deps: None)

    import pactown.sandbox_manager as sm_module

    captured: dict[str, object] = {"npm": 0}

    def fake_run(cmd, capture_output=False, text=False, check=False, env=None, cwd=None, **kwargs):
        if isinstance(cmd, list) and cmd and str(cmd[0]) == "npm":
            captured["npm"] = int(captured["npm"]) + 1
        return SimpleNamespace(returncode=0, stdout="", stderr="")

    monkeypatch.setattr(sm_module.subprocess, "run", fake_run)

    readme = """# Express Hello

```javascript markpact:file path=index.js
console.log('hello');
```

```node markpact:deps
express
```

```bash markpact:run
node index.js $PORT
```
"""

    readme_path = tmp_path / "README.md"
    readme_path.write_text(readme)

    service = ServiceConfig(name="svc", readme=str(readme_path), port=8000)

    sandbox = manager.create_sandbox(
        service=service,
        readme_path=readme_path,
        install_dependencies=True,
        on_log=None,
        env=None,
    )

    pkg = (Path(sandbox.path) / "package.json")
    assert pkg.exists()
    assert "express" in pkg.read_text(encoding="utf-8")
    assert int(captured["npm"]) == 1



"""
CQRS/Event Sourcing infrastructure for Pactown.

Provides a complete event sourcing foundation:
- Event: Immutable event records with versioning
- EventStore: Append-only event storage with subscriptions
- Aggregate: Base class for event-sourced aggregates
- Projections: Materialized views built from events
- Commands/Queries: CQRS pattern implementation

Usage:
    from pactown.events import (
        Event, EventType, EventStore, get_event_store,
        ServiceAggregate, ServiceCommands, ServiceQueries,
    )
    
    # Record events
    commands = ServiceCommands(get_event_store())
    await commands.create_service(service_id=1, user_id=1, name="api", port=8000)
    await commands.start_service(service_id=1, pid=12345)
    
    # Query events
    queries = ServiceQueries(get_event_store())
    history = queries.get_service_history(service_id=1)
    stats = queries.get_stats()
"""
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, UTC
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar
import asyncio
import json
import uuid


class EventType(str, Enum):
    """Standard event types for service lifecycle."""
    # Service lifecycle
    SERVICE_CREATED = "service.created"
    SERVICE_STARTED = "service.started"
    SERVICE_STOPPED = "service.stopped"
    SERVICE_DELETED = "service.deleted"
    SERVICE_HEALTH_CHECK = "service.health_check"
    SERVICE_ERROR = "service.error"
    SERVICE_RESTARTED = "service.restarted"
    
    # Sandbox lifecycle
    SANDBOX_CREATED = "sandbox.created"
    SANDBOX_DESTROYED = "sandbox.destroyed"
    SANDBOX_FILES_WRITTEN = "sandbox.files_written"
    SANDBOX_DEPS_INSTALLED = "sandbox.deps_installed"
    
    # Project lifecycle
    PROJECT_CREATED = "project.created"
    PROJECT_UPDATED = "project.updated"
    PROJECT_DELETED = "project.deleted"
    PROJECT_VALIDATED = "project.validated"
    
    # User actions
    USER_LOGIN = "user.login"
    USER_LOGOUT = "user.logout"
    USER_CREATED = "user.created"
    
    # Security events
    SECURITY_CHECK_PASSED = "security.check_passed"
    SECURITY_CHECK_FAILED = "security.check_failed"
    RATE_LIMIT_HIT = "security.rate_limit"
    ANOMALY_DETECTED = "security.anomaly"
    
    # Deployment events
    DEPLOYMENT_STARTED = "deployment.started"
    DEPLOYMENT_COMPLETED = "deployment.completed"
    DEPLOYMENT_FAILED = "deployment.failed"
    
    # Custom events
    CUSTOM = "custom"


@dataclass(frozen=True)
class Event:
    """
    Immutable event record.
    
    Events are the source of truth in event sourcing. Each event represents
    a fact that happened in the system at a specific point in time.
    
    Attributes:
        event_type: The type of event (from EventType enum or custom string)
        aggregate_id: ID of the aggregate this event belongs to (e.g., "service:123")
        aggregate_type: Type of aggregate (e.g., "service", "project", "user")
        data: Event payload with domain-specific data
        metadata: Additional context (user_id, correlation_id, etc.)
        timestamp: When the event occurred (UTC)
        event_id: Unique identifier for this event
        version: Event schema version for migrations
        sequence: Position in the event stream (set by EventStore)
    """
    event_type: EventType | str
    aggregate_id: str
    aggregate_type: str
    data: Dict[str, Any]
    metadata: Dict[str, Any] = field(default_factory=dict)
    timestamp: datetime = field(default_factory=lambda: datetime.now(UTC))
    event_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    version: int = 1
    sequence: int = 0
    
    def to_dict(self) -> Dict[str, Any]:
        """Serialize event to dictionary for JSON storage."""
        return {
            "event_id": self.event_id,
            "event_type": self.event_type.value if isinstance(self.event_type, EventType) else self.event_type,
            "aggregate_id": self.aggregate_id,
            "aggregate_type": self.aggregate_type,
            "data": self.data,
            "metadata": self.metadata,
            "timestamp": self.timestamp.isoformat(),
            "version": self.version,
            "sequence": self.sequence,
        }
    
    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Event":
        """Deserialize event from dictionary."""
        event_type_str = d["event_type"]
        try:
            event_type = EventType(event_type_str)
        except ValueError:
            event_type = event_type_str
        
        return cls(
            event_id=d["event_id"],
            event_type=event_type,
            aggregate_id=d["aggregate_id"],
            aggregate_type=d["aggregate_type"],
            data=d["data"],
            metadata=d.get("metadata", {}),
            timestamp=datetime.fromisoformat(d["timestamp"]) if isinstance(d["timestamp"], str) else d["timestamp"],
            version=d.get("version", 1),
            sequence=d.get("sequence", 0),
        )


class EventStore:
    """
    Append-only event store with subscription support.
    
    Provides:
    - Append-only event storage
    - Event subscriptions for reactive updates
    - Querying by aggregate, type, or time range
    - Optional persistence to JSON file
    
    Thread-safe for async operations.
    """
    
    def __init__(self, persistence_path: Optional[Path] = None):
        """
        Initialize event store.
        
        Args:
            persistence_path: Optional path to persist events to JSON file
        """
        self._events: List[Event] = []
        self._subscribers: Dict[EventType | str, List[Callable]] = defaultdict(list)
        self._global_subscribers: List[Callable] = []
        self._lock = asyncio.Lock()
        self._sequence = 0
        self._persistence_path = persistence_path
        
        if persistence_path and persistence_path.exists():
            self._load_from_file()
    
    def _load_from_file(self) -> None:
        """Load events from persistence file."""
        try:
            with open(self._persistence_path, 'r') as f:
                data = json.load(f)
                self._events = [Event.from_dict(e) for e in data.get("events", [])]
                self._sequence = data.get("sequence", len(self._events))
        except (json.JSONDecodeError, KeyError):
            pass
    
    def _save_to_file(self) -> None:
        """Persist events to file."""
        if not self._persistence_path:
            return
        
        self._persistence_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._persistence_path, 'w') as f:
            json.dump({
                "events": [e.to_dict() for e in self._events],
                "sequence": self._sequence,
            }, f, indent=2, default=str)
    
    async def append(self, event: Event) -> Event:
        """
        Append event to store and notify subscribers.
        
        Args:
            event: Event to append
            
        Returns:
            Event with sequence number set
        """
        async with self._lock:
            self._sequence += 1
            # Create new event with sequence (Event is frozen)
            sequenced_event = Event(
                event_id=event.event_id,
                event_type=event.event_type,
                aggregate_id=event.aggregate_id,
                aggregate_type=event.aggregate_type,
                data=event.data,
                metadata=event.metadata,
                timestamp=event.timestamp,
                version=event.version,
                sequence=self._sequence,
            )
            self._events.append(sequenced_event)
            
            if self._persistence_path:
                self._save_to_file()
        
        # Notify subscribers asynchronously
        await self._notify_subscribers(sequenced_event)
        
        return sequenced_event
    
    async def _notify_subscribers(self, event: Event) -> None:
        """Notify all relevant subscribers of an event."""
        handlers = (
            self._subscribers.get(event.event_type, []) +
            self._global_subscribers
        )
        
        for handler in handlers:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(event)
                else:
                    handler(event)
            except Exception as e:
                print(f"Event handler error: {e}")
    
    def subscribe(self, event_type: EventType | str, handler: Callable) -> Callable[[], None]:
        """
        Subscribe to events of a specific type.
        
        Args:
            event_type: Type of events to subscribe to
            handler: Callback function (sync or async)
            
        Returns:
            Unsubscribe function
        """
        self._subscribers[event_type].append(handler)
        
        def unsubscribe():
            self._subscribers[event_type].remove(handler)
        
        return unsubscribe
    
    def subscribe_all(self, handler: Callable) -> Callable[[], None]:
        """
        Subscribe to all events.
        
        Args:
            handler: Callback function (sync or async)
            
        Returns:
            Unsubscribe function
        """
        self._global_subscribers.append(handler)
        
        def unsubscribe():
            self._global_subscribers.remove(handler)
        
        return unsubscribe
    
    def get_events(
        self,
        aggregate_id: Optional[str] = None,
        aggregate_type: Optional[str] = None,
        event_type: Optional[EventType | str] = None,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None,
        since_sequence: Optional[int] = None,
        limit: int = 100,
    ) -> List[Event]:
        """
        Query events with filters.
        
        Args:
            aggregate_id: Filter by aggregate ID
            aggregate_type: Filter by aggregate type
            event_type: Filter by event type
            since: Filter events after this timestamp
            until: Filter events before this timestamp
            since_sequence: Filter events after this sequence number
            limit: Maximum number of events to return
            
        Returns:
            List of matching events
        """
        events = self._events
        
        if aggregate_id:
            events = [e for e in events if e.aggregate_id == aggregate_id]
        if aggregate_type:
            events = [e for e in events if e.aggregate_type == aggregate_type]
        if event_type:
            events = [e for e in events if e.event_type == event_type]
        if since:
            events = [e for e in events if e.timestamp >= since]
        if until:
            events = [e for e in events if e.timestamp <= until]
        if since_sequence is not None:
            events = [e for e in events if e.sequence > since_sequence]
        
        return events[-limit:]
    
    def get_aggregate_history(self, aggregate_id: str) -> List[Event]:
        """Get all events for a specific aggregate in order."""
        return sorted(
            [e for e in self._events if e.aggregate_id == aggregate_id],
            key=lambda e: e.sequence
        )
    
    def count(self, event_type: Optional[EventType | str] = None) -> int:
        """Count events, optionally filtered by type."""
        if event_type:
            return len([e for e in self._events if e.event_type == event_type])
        return len(self._events)
    
    def get_current_sequence(self) -> int:
        """Get current sequence number."""
        return self._sequence
    
    def clear(self) -> None:
        """Clear all events (use with caution)."""
        self._events.clear()
        self._sequence = 0
        if self._persistence_path and self._persistence_path.exists():
            self._persistence_path.unlink()


# Global event store instance
_event_store: Optional[EventStore] = None


def get_event_store(persistence_path: Optional[Path] = None) -> EventStore:
    """Get or create global event store."""
    global _event_store
    if _event_store is None:
        _event_store = EventStore(persistence_path=persistence_path)
    return _event_store


def set_event_store(store: EventStore) -> None:
    """Set the global event store instance."""
    global _event_store
    _event_store = store


# Aggregate base class
T = TypeVar('T', bound='Aggregate')


class Aggregate(ABC):
    """
    Base class for event-sourced aggregates.
    
    Aggregates encapsulate domain logic and maintain consistency boundaries.
    State is rebuilt by replaying events.
    
    Usage:
        class ServiceAggregate(Aggregate):
            def __init__(self, aggregate_id: str):
                super().__init__(aggregate_id, "service")
                self.name = ""
                self.status = "pending"
            
            def apply_event(self, event: Event) -> None:
                if event.event_type == EventType.SERVICE_CREATED:
                    self.name = event.data["name"]
                    self.status = "created"
                elif event.event_type == EventType.SERVICE_STARTED:
                    self.status = "running"
    """
    
    def __init__(self, aggregate_id: str, aggregate_type: str):
        self.aggregate_id = aggregate_id
        self.aggregate_type = aggregate_type
        self.version = 0
        self._pending_events: List[Event] = []
    
    @abstractmethod
    def apply_event(self, event: Event) -> None:
        """Apply an event to update aggregate state."""
        pass
    
    def load_from_history(self, events: List[Event]) -> None:
        """Rebuild state from event history."""
        for event in events:
            self.apply_event(event)
            self.version = event.sequence
    
    def raise_event(self, event_type: EventType | str, data: Dict[str, Any], metadata: Optional[Dict[str, Any]] = None) -> Event:
        """
        Raise a new event from this aggregate.
        
        Args:
            event_type: Type of event
            data: Event payload
            metadata: Optional metadata
            
        Returns:
            The raised event (not yet persisted)
        """
        event = Event(
            event_type=event_type,
            aggregate_id=self.aggregate_id,
            aggregate_type=self.aggregate_type,
            data=data,
            metadata=metadata or {},
        )
        self._pending_events.append(event)
        self.apply_event(event)
        return event
    
    def get_pending_events(self) -> List[Event]:
        """Get events raised but not yet persisted."""
        return self._pending_events.copy()
    
    def clear_pending_events(self) -> None:
        """Clear pending events after persistence."""
        self._pending_events.clear()
    
    @classmethod
    async def load(cls: type[T], aggregate_id: str, event_store: EventStore) -> T:
        """Load aggregate from event store."""
        instance = cls(aggregate_id)
        events = event_store.get_aggregate_history(aggregate_id)
        instance.load_from_history(events)
        return instance


class ServiceAggregate(Aggregate):
    """
    Event-sourced aggregate for service lifecycle.
    
    Tracks service state through events, enabling:
    - Full audit trail of service changes
    - State reconstruction at any point in time
    - Eventual consistency with read models
    """
    
    def __init__(self, aggregate_id: str):
        super().__init__(aggregate_id, "service")
        self.service_id: Optional[int] = None
        self.user_id: Optional[int] = None
        self.name: str = ""
        self.port: int = 0
        self.status: str = "pending"
        self.pid: Optional[int] = None
        self.started_at: Optional[datetime] = None
        self.stopped_at: Optional[datetime] = None
        self.error_count: int = 0
        self.last_error: Optional[str] = None
    
    def apply_event(self, event: Event) -> None:
        """Apply event to update service state."""
        if event.event_type == EventType.SERVICE_CREATED:
            self.service_id = event.data.get("service_id")
            self.user_id = event.data.get("user_id")
            self.name = event.data.get("name", "")
            self.port = event.data.get("port", 0)
            self.status = "created"
            
        elif event.event_type == EventType.SERVICE_STARTED:
            self.status = "running"
            self.pid = event.data.get("pid")
            self.started_at = event.timestamp
            
        elif event.event_type == EventType.SERVICE_STOPPED:
            self.status = "stopped"
            self.pid = None
            self.stopped_at = event.timestamp
            
        elif event.event_type == EventType.SERVICE_ERROR:
            self.error_count += 1
            self.last_error = event.data.get("error")
            if event.data.get("fatal", False):
                self.status = "error"
                
        elif event.event_type == EventType.SERVICE_DELETED:
            self.status = "deleted"
    
    def to_dict(self) -> Dict[str, Any]:
        """Serialize aggregate state."""
        return {
            "aggregate_id": self.aggregate_id,
            "service_id": self.service_id,
            "user_id": self.user_id,
            "name": self.name,
            "port": self.port,
            "status": self.status,
            "pid": self.pid,
            "started_at": self.started_at.isoformat() if self.started_at else None,
            "stopped_at": self.stopped_at.isoformat() if self.stopped_at else None,
            "error_count": self.error_count,
            "last_error": self.last_error,
            "version": self.version,
        }


# Command handlers (Write side of CQRS)
class ServiceCommands:
    """
    Command handlers for service operations.
    
    Commands represent intentions to change state. Each command
    results in one or more events being recorded.
    """
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    async def create_service(
        self,
        service_id: int,
        user_id: int,
        name: str,
        port: int,
        **kwargs
    ) -> Event:
        """Record service creation."""
        event = Event(
            event_type=EventType.SERVICE_CREATED,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={
                "service_id": service_id,
                "user_id": user_id,
                "name": name,
                "port": port,
                **kwargs,
            },
            metadata={"user_id": user_id},
        )
        return await self.event_store.append(event)
    
    async def start_service(
        self,
        service_id: int,
        pid: Optional[int] = None,
        startup_time_ms: Optional[float] = None,
        cached: bool = False,
    ) -> Event:
        """Record service start."""
        event = Event(
            event_type=EventType.SERVICE_STARTED,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={
                "service_id": service_id,
                "pid": pid,
                "startup_time_ms": startup_time_ms,
                "cached": cached,
            },
        )
        return await self.event_store.append(event)
    
    async def stop_service(
        self,
        service_id: int,
        reason: str = "user_request",
    ) -> Event:
        """Record service stop."""
        event = Event(
            event_type=EventType.SERVICE_STOPPED,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={
                "service_id": service_id,
                "reason": reason,
            },
        )
        return await self.event_store.append(event)
    
    async def record_error(
        self,
        service_id: int,
        error: str,
        details: Optional[Dict] = None,
        fatal: bool = False,
    ) -> Event:
        """Record service error."""
        event = Event(
            event_type=EventType.SERVICE_ERROR,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={
                "service_id": service_id,
                "error": error,
                "details": details or {},
                "fatal": fatal,
            },
        )
        return await self.event_store.append(event)
    
    async def record_health_check(
        self,
        service_id: int,
        healthy: bool,
        response_time_ms: Optional[float] = None,
        status_code: Optional[int] = None,
    ) -> Event:
        """Record health check result."""
        event = Event(
            event_type=EventType.SERVICE_HEALTH_CHECK,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={
                "service_id": service_id,
                "healthy": healthy,
                "response_time_ms": response_time_ms,
                "status_code": status_code,
            },
        )
        return await self.event_store.append(event)
    
    async def delete_service(
        self,
        service_id: int,
        user_id: Optional[int] = None,
    ) -> Event:
        """Record service deletion."""
        event = Event(
            event_type=EventType.SERVICE_DELETED,
            aggregate_id=f"service:{service_id}",
            aggregate_type="service",
            data={"service_id": service_id},
            metadata={"user_id": user_id} if user_id else {},
        )
        return await self.event_store.append(event)


class ProjectCommands:
    """Command handlers for project operations."""
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    async def create_project(
        self,
        project_id: int,
        user_id: int,
        name: str,
        **kwargs
    ) -> Event:
        """Record project creation."""
        event = Event(
            event_type=EventType.PROJECT_CREATED,
            aggregate_id=f"project:{project_id}",
            aggregate_type="project",
            data={
                "project_id": project_id,
                "user_id": user_id,
                "name": name,
                **kwargs,
            },
            metadata={"user_id": user_id},
        )
        return await self.event_store.append(event)
    
    async def update_project(
        self,
        project_id: int,
        changes: Dict[str, Any],
        user_id: Optional[int] = None,
    ) -> Event:
        """Record project update."""
        event = Event(
            event_type=EventType.PROJECT_UPDATED,
            aggregate_id=f"project:{project_id}",
            aggregate_type="project",
            data={
                "project_id": project_id,
                "changes": changes,
            },
            metadata={"user_id": user_id} if user_id else {},
        )
        return await self.event_store.append(event)
    
    async def delete_project(
        self,
        project_id: int,
        user_id: Optional[int] = None,
    ) -> Event:
        """Record project deletion."""
        event = Event(
            event_type=EventType.PROJECT_DELETED,
            aggregate_id=f"project:{project_id}",
            aggregate_type="project",
            data={"project_id": project_id},
            metadata={"user_id": user_id} if user_id else {},
        )
        return await self.event_store.append(event)


class SecurityCommands:
    """Command handlers for security events."""
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    async def record_security_check(
        self,
        user_id: str,
        service_id: str,
        passed: bool,
        reason: Optional[str] = None,
        details: Optional[Dict] = None,
    ) -> Event:
        """Record security check result."""
        event_type = EventType.SECURITY_CHECK_PASSED if passed else EventType.SECURITY_CHECK_FAILED
        event = Event(
            event_type=event_type,
            aggregate_id=f"user:{user_id}",
            aggregate_type="security",
            data={
                "user_id": user_id,
                "service_id": service_id,
                "passed": passed,
                "reason": reason,
                "details": details or {},
            },
        )
        return await self.event_store.append(event)
    
    async def record_rate_limit(
        self,
        user_id: str,
        endpoint: str,
        limit: int,
    ) -> Event:
        """Record rate limit hit."""
        event = Event(
            event_type=EventType.RATE_LIMIT_HIT,
            aggregate_id=f"user:{user_id}",
            aggregate_type="security",
            data={
                "user_id": user_id,
                "endpoint": endpoint,
                "limit": limit,
            },
        )
        return await self.event_store.append(event)
    
    async def record_anomaly(
        self,
        user_id: str,
        anomaly_type: str,
        severity: str,
        details: Optional[Dict] = None,
    ) -> Event:
        """Record security anomaly."""
        event = Event(
            event_type=EventType.ANOMALY_DETECTED,
            aggregate_id=f"user:{user_id}",
            aggregate_type="security",
            data={
                "user_id": user_id,
                "anomaly_type": anomaly_type,
                "severity": severity,
                "details": details or {},
            },
        )
        return await self.event_store.append(event)


# Query handlers (Read side of CQRS)
class ServiceQueries:
    """
    Query handlers for service read operations.
    
    Queries don't modify state - they only read from the event store
    or materialized projections.
    """
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    def get_service_history(self, service_id: int) -> List[Dict]:
        """Get history of events for a service."""
        events = self.event_store.get_aggregate_history(f"service:{service_id}")
        return [e.to_dict() for e in events]
    
    def get_recent_starts(self, limit: int = 10) -> List[Dict]:
        """Get recent service starts."""
        events = self.event_store.get_events(
            event_type=EventType.SERVICE_STARTED,
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_recent_errors(self, limit: int = 10) -> List[Dict]:
        """Get recent service errors."""
        events = self.event_store.get_events(
            event_type=EventType.SERVICE_ERROR,
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_recent_health_checks(self, service_id: Optional[int] = None, limit: int = 10) -> List[Dict]:
        """Get recent health check results."""
        aggregate_id = f"service:{service_id}" if service_id else None
        events = self.event_store.get_events(
            aggregate_id=aggregate_id,
            event_type=EventType.SERVICE_HEALTH_CHECK,
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_stats(self) -> Dict[str, int]:
        """Get event statistics."""
        return {
            "total_events": self.event_store.count(),
            "services_created": self.event_store.count(EventType.SERVICE_CREATED),
            "services_started": self.event_store.count(EventType.SERVICE_STARTED),
            "services_stopped": self.event_store.count(EventType.SERVICE_STOPPED),
            "services_deleted": self.event_store.count(EventType.SERVICE_DELETED),
            "errors": self.event_store.count(EventType.SERVICE_ERROR),
            "health_checks": self.event_store.count(EventType.SERVICE_HEALTH_CHECK),
        }
    
    async def get_service_state(self, service_id: int) -> Dict[str, Any]:
        """Rebuild current service state from events."""
        aggregate = await ServiceAggregate.load(
            f"service:{service_id}",
            self.event_store
        )
        return aggregate.to_dict()
    
    def get_user_services(self, user_id: int) -> List[Dict]:
        """Get all service events for a user."""
        events = self.event_store.get_events(
            event_type=EventType.SERVICE_CREATED,
        )
        user_events = [e for e in events if e.data.get("user_id") == user_id]
        return [e.to_dict() for e in user_events]


class ProjectQueries:
    """Query handlers for project read operations."""
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    def get_project_history(self, project_id: int) -> List[Dict]:
        """Get history of events for a project."""
        events = self.event_store.get_aggregate_history(f"project:{project_id}")
        return [e.to_dict() for e in events]
    
    def get_recent_projects(self, user_id: Optional[int] = None, limit: int = 10) -> List[Dict]:
        """Get recently created projects."""
        events = self.event_store.get_events(
            event_type=EventType.PROJECT_CREATED,
            limit=limit,
        )
        if user_id:
            events = [e for e in events if e.data.get("user_id") == user_id]
        return [e.to_dict() for e in events]
    
    def get_stats(self) -> Dict[str, int]:
        """Get project statistics."""
        return {
            "projects_created": self.event_store.count(EventType.PROJECT_CREATED),
            "projects_updated": self.event_store.count(EventType.PROJECT_UPDATED),
            "projects_deleted": self.event_store.count(EventType.PROJECT_DELETED),
        }


class SecurityQueries:
    """Query handlers for security read operations."""
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
    
    def get_recent_security_failures(self, limit: int = 10) -> List[Dict]:
        """Get recent security check failures."""
        events = self.event_store.get_events(
            event_type=EventType.SECURITY_CHECK_FAILED,
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_user_security_history(self, user_id: str, limit: int = 50) -> List[Dict]:
        """Get security event history for a user."""
        events = self.event_store.get_events(
            aggregate_id=f"user:{user_id}",
            aggregate_type="security",
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_rate_limit_hits(self, since: Optional[datetime] = None, limit: int = 100) -> List[Dict]:
        """Get recent rate limit hits."""
        events = self.event_store.get_events(
            event_type=EventType.RATE_LIMIT_HIT,
            since=since,
            limit=limit,
        )
        return [e.to_dict() for e in events]
    
    def get_anomalies(self, severity: Optional[str] = None, limit: int = 50) -> List[Dict]:
        """Get security anomalies."""
        events = self.event_store.get_events(
            event_type=EventType.ANOMALY_DETECTED,
            limit=limit,
        )
        if severity:
            events = [e for e in events if e.data.get("severity") == severity]
        return [e.to_dict() for e in events]
    
    def get_stats(self) -> Dict[str, int]:
        """Get security statistics."""
        return {
            "security_checks_passed": self.event_store.count(EventType.SECURITY_CHECK_PASSED),
            "security_checks_failed": self.event_store.count(EventType.SECURITY_CHECK_FAILED),
            "rate_limit_hits": self.event_store.count(EventType.RATE_LIMIT_HIT),
            "anomalies_detected": self.event_store.count(EventType.ANOMALY_DETECTED),
        }


# Projections - Materialized views built from events
class Projection(ABC):
    """
    Base class for event projections.
    
    Projections maintain materialized views that are optimized for
    specific query patterns. They're rebuilt by replaying events.
    """
    
    def __init__(self, event_store: EventStore):
        self.event_store = event_store
        self._last_sequence = 0
    
    @abstractmethod
    def apply(self, event: Event) -> None:
        """Apply an event to update the projection."""
        pass
    
    def rebuild(self) -> None:
        """Rebuild projection from all events."""
        self._last_sequence = 0
        for event in self.event_store.get_events(limit=10000):
            self.apply(event)
            self._last_sequence = event.sequence
    
    def catch_up(self) -> None:
        """Apply new events since last update."""
        events = self.event_store.get_events(
            since_sequence=self._last_sequence,
            limit=1000,
        )
        for event in events:
            self.apply(event)
            self._last_sequence = event.sequence


class ServiceStatusProjection(Projection):
    """
    Projection maintaining current status of all services.
    
    Optimized for queries like "list all running services".
    """
    
    def __init__(self, event_store: EventStore):
        super().__init__(event_store)
        self._services: Dict[str, Dict[str, Any]] = {}
    
    def apply(self, event: Event) -> None:
        """Update service status based on event."""
        if event.aggregate_type != "service":
            return
        
        service_id = event.aggregate_id
        
        if event.event_type == EventType.SERVICE_CREATED:
            self._services[service_id] = {
                "service_id": event.data.get("service_id"),
                "user_id": event.data.get("user_id"),
                "name": event.data.get("name"),
                "port": event.data.get("port"),
                "status": "created",
                "created_at": event.timestamp.isoformat(),
            }
        elif event.event_type == EventType.SERVICE_STARTED:
            if service_id in self._services:
                self._services[service_id]["status"] = "running"
                self._services[service_id]["pid"] = event.data.get("pid")
                self._services[service_id]["started_at"] = event.timestamp.isoformat()
        elif event.event_type == EventType.SERVICE_STOPPED:
            if service_id in self._services:
                self._services[service_id]["status"] = "stopped"
                self._services[service_id]["pid"] = None
        elif event.event_type == EventType.SERVICE_DELETED:
            self._services.pop(service_id, None)
        elif event.event_type == EventType.SERVICE_ERROR:
            if service_id in self._services:
                self._services[service_id]["last_error"] = event.data.get("error")
                if event.data.get("fatal"):
                    self._services[service_id]["status"] = "error"
    
    def get_all(self) -> List[Dict[str, Any]]:
        """Get all services."""
        return list(self._services.values())
    
    def get_running(self) -> List[Dict[str, Any]]:
        """Get only running services."""
        return [s for s in self._services.values() if s.get("status") == "running"]
    
    def get_by_user(self, user_id: int) -> List[Dict[str, Any]]:
        """Get services for a specific user."""
        return [s for s in self._services.values() if s.get("user_id") == user_id]
    
    def get(self, service_id: str) -> Optional[Dict[str, Any]]:
        """Get a specific service."""
        return self._services.get(service_id)


# Convenience functions for common patterns
def get_service_commands(event_store: Optional[EventStore] = None) -> ServiceCommands:
    """Get service command handlers."""
    return ServiceCommands(event_store or get_event_store())


def get_service_queries(event_store: Optional[EventStore] = None) -> ServiceQueries:
    """Get service query handlers."""
    return ServiceQueries(event_store or get_event_store())


def get_project_commands(event_store: Optional[EventStore] = None) -> ProjectCommands:
    """Get project command handlers."""
    return ProjectCommands(event_store or get_event_store())


def get_project_queries(event_store: Optional[EventStore] = None) -> ProjectQueries:
    """Get project query handlers."""
    return ProjectQueries(event_store or get_event_store())


def get_security_commands(event_store: Optional[EventStore] = None) -> SecurityCommands:
    """Get security command handlers."""
    return SecurityCommands(event_store or get_event_store())


def get_security_queries(event_store: Optional[EventStore] = None) -> SecurityQueries:
    """Get security query handlers."""
    return SecurityQueries(event_store or get_event_store())



"""Sandbox manager for pactown services."""

import json
import logging
import os
import inspect
import re
import shutil
import stat
import signal
import subprocess
import tempfile
import time
import socket
import shlex
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime, UTC
from pathlib import Path
from threading import Event
from threading import Thread
from threading import Lock
from typing import Callable, Optional, List, Dict, Any

from markpact import Sandbox, ensure_venv

from .config import ServiceConfig
from .markpact_blocks import parse_blocks
from .fast_start import DependencyCache

# Configure detailed logging
logger = logging.getLogger("pactown.sandbox")
logger.setLevel(logging.DEBUG)


def _ui_log_level() -> int:
    raw = str(os.environ.get("PACTOWN_UI_LOG_LEVEL", "INFO") or "INFO").strip().upper()
    if raw == "DEBUG":
        return logging.DEBUG
    if raw == "WARNING" or raw == "WARN":
        return logging.WARNING
    if raw == "ERROR":
        return logging.ERROR
    if raw == "CRITICAL":
        return logging.CRITICAL
    return logging.INFO


def _should_emit_to_ui(level: str) -> bool:
    try:
        lvl = int(getattr(logging, str(level).upper()))
    except Exception:
        lvl = logging.INFO
    return lvl >= _ui_log_level()


_ON_LOG_ACCEPTS_LEVEL: dict[int, bool] = {}


def _call_on_log(on_log: Optional[Callable[..., None]], msg: str, level: str) -> None:
    if not on_log:
        return
    key = id(on_log)
    accepts = _ON_LOG_ACCEPTS_LEVEL.get(key)
    if accepts is None:
        try:
            sig = inspect.signature(on_log)
            params = list(sig.parameters.values())
            accepts = any(p.kind == inspect.Parameter.VAR_POSITIONAL for p in params) or len(params) >= 2
        except Exception:
            accepts = False
        _ON_LOG_ACCEPTS_LEVEL[key] = accepts
    if accepts:
        on_log(msg, level)
    else:
        on_log(msg)


def _heartbeat(
    *,
    stop: Event,
    on_log: Optional[Callable[..., None]],
    message: str,
    interval_s: float = 1.0,
) -> None:
    if not on_log:
        return
    started = time.monotonic()
    ticks = 0
    while not stop.wait(interval_s):
        ticks += 1
        elapsed = int(time.monotonic() - started)
        if _should_emit_to_ui("INFO"):
            _call_on_log(on_log, f"⏳ {message} (elapsed={elapsed}s)", "INFO")


def _beat_every_s(*, default: int = 5) -> int:
    try:
        return max(1, int(os.environ.get("PACTOWN_HEALTH_HEARTBEAT_S", str(default))))
    except Exception:
        return default

# File handler for persistent logs
LOG_DIR = Path(os.environ.get("PACTOWN_LOG_DIR", tempfile.gettempdir() + "/pactown-logs"))
LOG_DIR.mkdir(parents=True, exist_ok=True)
_log_path = str(LOG_DIR / "sandbox.log")
if not any(
    isinstance(h, logging.FileHandler) and getattr(h, "baseFilename", None) == _log_path
    for h in logger.handlers
):
    file_handler = logging.FileHandler(_log_path)
    file_handler.setFormatter(logging.Formatter(
        '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
    ))
    logger.addHandler(file_handler)


def _path_debug(path: Path) -> str:
    try:
        st = path.stat() if path.exists() else None
        mode = oct(st.st_mode & 0o777) if st else "-"
        uid = st.st_uid if st else "-"
        gid = st.st_gid if st else "-"
    except Exception:
        mode, uid, gid = "?", "?", "?"
    try:
        readable = os.access(path, os.R_OK)
        writable = os.access(path, os.W_OK)
        executable = os.access(path, os.X_OK)
    except Exception:
        readable, writable, executable = False, False, False
    return (
        f"path={path} exists={path.exists()} is_dir={path.is_dir()} is_file={path.is_file()} "
        f"mode={mode} uid={uid} gid={gid} access=r{int(readable)}w{int(writable)}x{int(executable)}"
    )


def _escape_dotenv_value(value: str) -> str:
    v = str(value)
    v = v.replace("\\", "\\\\")
    v = v.replace("\n", "\\n")
    v = v.replace("\r", "\\r")
    v = v.replace('"', '\\"')
    return f'"{v}"'


def _write_dotenv_file(sandbox_path: Path, env: dict[str, str]) -> None:
    lines: list[str] = []
    for key, value in (env or {}).items():
        if value is None:
            continue
        if not isinstance(key, str):
            continue
        if not re.match(r"^[A-Za-z_][A-Za-z0-9_]*$", key):
            continue
        lines.append(f"{key}={_escape_dotenv_value(str(value))}")

    dotenv_path = sandbox_path / ".env"
    dotenv_path.write_text("\n".join(lines) + ("\n" if lines else ""))
    try:
        dotenv_path.chmod(0o600)
    except Exception:
        pass


def _sandbox_fallback_ids() -> tuple[int, int]:
    try:
        uid = int(os.environ.get("PACTOWN_SANDBOX_UID", "65534"))
    except Exception:
        uid = 65534
    try:
        gid = int(os.environ.get("PACTOWN_SANDBOX_GID", str(uid)))
    except Exception:
        gid = uid
    return uid, gid


def _chown_sandbox_tree(sandbox_path: Path, uid: int, gid: int) -> None:
    try:
        os.chown(sandbox_path, uid, gid)
    except Exception:
        pass
    try:
        sandbox_path.chmod(0o700)
    except Exception:
        pass

    for root, dirnames, filenames in os.walk(sandbox_path):
        if ".venv" in dirnames:
            dirnames.remove(".venv")

        root_path = Path(root)
        try:
            os.chown(root_path, uid, gid)
        except Exception:
            pass
        try:
            root_path.chmod(0o700)
        except Exception:
            pass

        for name in filenames:
            p = root_path / name
            try:
                st = os.lstat(p)
                if stat.S_ISLNK(st.st_mode):
                    continue
            except Exception:
                continue
            try:
                os.chown(p, uid, gid)
            except Exception:
                pass


@dataclass
class ServiceProcess:
    """Represents a running service process."""
    name: str
    pid: int
    port: Optional[int]
    sandbox_path: Path
    process: Optional[subprocess.Popen] = None
    started_at: float = field(default_factory=time.time)

    @property
    def is_running(self) -> bool:
        if self.process:
            return self.process.poll() is None
        try:
            os.kill(self.pid, 0)
            if self.port:
                try:
                    with socket.create_connection(("127.0.0.1", int(self.port)), timeout=0.2):
                        return True
                except OSError:
                    return False
            return True
        except OSError:
            return False


class SandboxManager:
    """Manages sandboxes for multiple services."""

    @staticmethod
    def _is_node_lang(lang: str) -> bool:
        l = (lang or "").strip().lower()
        return l in {"node", "js", "javascript", "npm"}

    @staticmethod
    def _infer_node_project(*, blocks: list, deps: list[str], run_cmd: str) -> bool:
        for b in blocks:
            if getattr(b, "kind", "") == "deps" and SandboxManager._is_node_lang(getattr(b, "lang", "")):
                return True
            if getattr(b, "kind", "") == "file":
                try:
                    p = str(b.get_path() or "").strip()
                except Exception:
                    p = ""
                if p:
                    pl = p.lower()
                    if pl == "package.json" or pl.endswith((".js", ".mjs", ".cjs", ".ts", ".tsx", ".jsx")):
                        return True
        rc = (run_cmd or "").strip().lower()
        if rc.startswith("node ") or rc.startswith("npm ") or rc.startswith("pnpm ") or rc.startswith("yarn "):
            return True
        if " node " in f" {rc} ":
            return True
        deps_l = {d.strip().lower() for d in (deps or []) if str(d).strip()}
        if "express" in deps_l and ("node" in rc or "npm" in rc):
            return True
        return False

    @staticmethod
    def _ensure_package_json(*, sandbox_path: Path, service_name: str, deps: list[str]) -> None:
        pkg_path = sandbox_path / "package.json"
        if pkg_path.exists():
            return

        deps_obj: dict[str, str] = {}
        for dep in deps:
            d = str(dep).strip()
            if not d:
                continue
            if d.startswith("#"):
                continue
            if " " in d:
                d = d.split()[0]
            if "@" in d and not d.startswith("@"):  # e.g. express@4
                name, version = d.rsplit("@", 1)
                name = name.strip()
                version = version.strip()
                if name:
                    deps_obj[name] = version or "latest"
            else:
                deps_obj[d] = "latest"

        pkg_path.write_text(
            json.dumps(
                {
                    "name": re.sub(r"[^a-z0-9-_]", "-", str(service_name).lower()) or "pactown-app",
                    "version": "1.0.0",
                    "private": True,
                    "dependencies": deps_obj,
                },
                indent=2,
                sort_keys=True,
            )
            + "\n",
            encoding="utf-8",
        )

    def _install_node_deps(
        self,
        *,
        sandbox: Sandbox,
        deps: list[str],
        on_log: Optional[Callable[[str], None]] = None,
        env: Optional[dict[str, str]] = None,
    ) -> None:
        def dbg(msg: str, level: str = "DEBUG"):
            logger.log(getattr(logging, level), f"[{sandbox.path.name}] {msg}")
            if on_log and _should_emit_to_ui(level):
                _call_on_log(on_log, msg, level)

        deps_clean = [str(d).strip() for d in (deps or []) if str(d).strip()]
        if not deps_clean:
            return

        dbg("Installing dependencies via npm", "INFO")
        stop = Event()
        thr = Thread(
            target=_heartbeat,
            kwargs={
                "stop": stop,
                "on_log": on_log,
                "message": f"[deploy] Installing dependencies via npm ({len(deps_clean)} deps)",
                "interval_s": float(_beat_every_s()),
            },
            daemon=True,
        )
        thr.start()
        try:
            install_env = os.environ.copy()
            for k, v in (env or {}).items():
                if k is None or v is None:
                    continue
                install_env[str(k)] = str(v)

            proc = subprocess.Popen(
                [
                    "npm",
                    "install",
                    "--no-audit",
                    "--no-fund",
                    "--progress=false",
                ],
                cwd=str(sandbox.path),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1,
                env=install_env,
            )
            try:
                if proc.stdout:
                    for line in proc.stdout:
                        s = (line or "").rstrip("\n")
                        if not s:
                            continue
                        if on_log and _should_emit_to_ui("INFO"):
                            _call_on_log(on_log, s, "INFO")
                rc = proc.wait()
                if rc != 0:
                    raise subprocess.CalledProcessError(rc, proc.args)
            finally:
                try:
                    if proc.stdout:
                        proc.stdout.close()
                except Exception:
                    pass
        except FileNotFoundError as e:
            dbg("npm not found in PATH (Node.js runtime missing)", "ERROR")
            raise e
        except subprocess.CalledProcessError as e:
            dbg(f"npm install failed: {e}", "ERROR")
            raise
        finally:
            stop.set()
        dbg("Dependencies installed", "INFO")

    def __init__(self, sandbox_root: str | Path):
        self.sandbox_root = Path(sandbox_root)
        self.sandbox_root.mkdir(parents=True, exist_ok=True)
        self._processes: dict[str, ServiceProcess] = {}
        self._dep_cache = DependencyCache(self.sandbox_root / ".cache" / "venvs")

    def get_sandbox_path(self, service_name: str) -> Path:
        """Get sandbox path for a service."""
        return self.sandbox_root / service_name

    def create_sandbox(
        self,
        service: ServiceConfig,
        readme_path: Path,
        install_dependencies: bool = True,
        on_log: Optional[Callable[[str], None]] = None,
        env: Optional[dict[str, str]] = None,
    ) -> Sandbox:
        """Create a sandbox for a service from its README."""
        def dbg(msg: str, level: str = "DEBUG"):
            logger.log(getattr(logging, level), f"[{service.name}] {msg}")
            if on_log and _should_emit_to_ui(level):
                _call_on_log(on_log, msg, level)

        def _verify_restored_venv(*, venv_path: Path, deps: list[str], run_cmd: str) -> bool:
            py = venv_path / "bin" / "python"
            if not py.exists():
                return False

            deps_l = {d.strip().lower() for d in deps if d.strip()}
            rc = (run_cmd or "").strip().lower()

            imports: list[str] = []
            if rc.startswith("uvicorn ") or " uvicorn " in f" {rc} ":
                imports.extend(["uvicorn", "click"])
            elif rc.startswith("gunicorn ") or " gunicorn " in f" {rc} ":
                imports.append("gunicorn")

            if "fastapi" in deps_l and "fastapi" not in imports:
                imports.append("fastapi")
            if "flask" in deps_l and "flask" not in imports:
                imports.append("flask")

            if not imports:
                return True

            code = "import importlib\n" + "\n".join([f"importlib.import_module({m!r})" for m in imports])

            try:
                check_env = os.environ.copy()
                for k, v in (env or {}).items():
                    if k is None or v is None:
                        continue
                    check_env[str(k)] = str(v)
                res = subprocess.run(
                    [str(py), "-c", code],
                    capture_output=True,
                    text=True,
                    env=check_env,
                    timeout=20,
                )
            except (FileNotFoundError, subprocess.TimeoutExpired):
                return False

            if res.returncode != 0:
                out = ((res.stderr or "") + "\n" + (res.stdout or "")).strip()[:2000]
                if out:
                    dbg(f"Cached venv verification failed: {out}", "WARNING")
                return False

            return True

        sandbox_path = self.get_sandbox_path(service.name)

        dbg(f"Sandbox root: {_path_debug(self.sandbox_root)}", "DEBUG")
        dbg(f"Sandbox path: {_path_debug(sandbox_path)}", "DEBUG")
        dbg(f"README path: {_path_debug(readme_path)}", "DEBUG")
        dbg(f"UID/EUID/GID: uid={os.getuid()} euid={os.geteuid()} gid={os.getgid()}", "DEBUG")

        if sandbox_path.exists():
            dbg(f"Removing existing sandbox: {sandbox_path}", "INFO")
            shutil.rmtree(sandbox_path)
        sandbox_path.mkdir(parents=True, exist_ok=False)
        dbg(f"Created sandbox dir: {_path_debug(sandbox_path)}", "DEBUG")

        sandbox = Sandbox(sandbox_path)

        readme_content = readme_path.read_text()
        dbg(f"Read README bytes={len(readme_content.encode('utf-8', errors='replace'))}", "DEBUG")
        blocks = parse_blocks(readme_content)

        kind_counts: dict[str, int] = {}
        for b in blocks:
            kind_counts[b.kind] = kind_counts.get(b.kind, 0) + 1
        dbg(f"Parsed markpact blocks: total={len(blocks)} kinds={kind_counts}", "DEBUG")

        deps: list[str] = []
        deps_node: list[str] = []
        run_cmd: str = ""

        for block in blocks:
            if block.kind == "deps":
                if self._is_node_lang(getattr(block, "lang", "")):
                    deps_node.extend(block.body.strip().split("\n"))
                else:
                    deps.extend(block.body.strip().split("\n"))
            elif block.kind == "file":
                file_path = block.get_path() or "main.py"
                dbg(f"Writing file: {file_path} (chars={len(block.body)})", "DEBUG")
                sandbox.write_file(file_path, block.body)
            elif block.kind == "run":
                run_cmd = block.body.strip()

        deps_clean = [d.strip() for d in deps if d.strip()]
        deps_node_clean = [d.strip() for d in deps_node if d.strip()]

        def _dep_name(raw: str) -> str:
            s = (raw or "").strip()
            if not s:
                return ""
            s = s.split(";")[0].strip()  # markers
            s = s.split("[")[0].strip()  # extras
            s = re.split(r"[<>=!~]", s, maxsplit=1)[0].strip()
            return s.lower()
        is_node = self._infer_node_project(blocks=blocks, deps=(deps_node_clean or deps_clean), run_cmd=run_cmd)
        effective_node_deps = deps_node_clean if deps_node_clean else (deps_clean if is_node else [])

        if is_node and effective_node_deps:
            dbg(f"Dependencies detected: count={len(effective_node_deps)}", "INFO")
            self._ensure_package_json(sandbox_path=sandbox.path, service_name=service.name, deps=effective_node_deps)
            dbg(f"Wrote package.json: {_path_debug(sandbox.path / 'package.json')}", "DEBUG")

            if install_dependencies:
                self._install_node_deps(sandbox=sandbox, deps=effective_node_deps, on_log=on_log, env=env)

            return sandbox

        if deps_clean:
            # Always write requirements.txt so the sandbox can be used as a container build context
            dbg(f"Dependencies detected: count={len(deps_clean)}", "INFO")

            run_l = (run_cmd or "").strip().lower()
            dep_names = {_dep_name(d) for d in deps_clean}
            if (run_l.startswith("uvicorn ") or " uvicorn " in f" {run_l} ") and "uvicorn" not in dep_names:
                deps_clean.append("uvicorn")
                dbg("Added implicit dependency: uvicorn (based on run command)", "INFO")
            if (run_l.startswith("gunicorn ") or " gunicorn " in f" {run_l} ") and "gunicorn" not in dep_names:
                deps_clean.append("gunicorn")
                dbg("Added implicit dependency: gunicorn (based on run command)", "INFO")

            sandbox.write_requirements(deps_clean)
            dbg(f"Wrote requirements.txt: {_path_debug(sandbox.path / 'requirements.txt')}", "DEBUG")

            if install_dependencies:
                cached = None
                try:
                    cached = self._dep_cache.get_cached_venv(deps_clean) if self._dep_cache else None
                except Exception:
                    cached = None

                if cached:
                    try:
                        dbg(f"⚡ Cache hit! Reusing venv ({cached.deps_hash})", "INFO")
                        venv_dst = sandbox.path / ".venv"
                        if venv_dst.exists() or venv_dst.is_symlink():
                            try:
                                if venv_dst.is_dir() and not venv_dst.is_symlink():
                                    shutil.rmtree(venv_dst)
                                else:
                                    venv_dst.unlink()
                            except Exception:
                                pass

                        def _copytree_fast(src_path: Path, dst_path: Path) -> None:
                            try:
                                shutil.copytree(src_path, dst_path, copy_function=os.link)
                            except Exception:
                                if dst_path.exists():
                                    shutil.rmtree(dst_path)
                                shutil.copytree(src_path, dst_path)

                        stop = Event()
                        thr = Thread(
                            target=_heartbeat,
                            kwargs={
                                "stop": stop,
                                "on_log": on_log,
                                "message": f"[deploy] Restoring cached venv ({len(deps_clean)} deps)",
                                "interval_s": float(_beat_every_s()),
                            },
                            daemon=True,
                        )
                        thr.start()
                        _copytree_fast(cached.path, venv_dst)
                        stop.set()
                        dbg(f"Venv restored: {_path_debug(venv_dst)}", "DEBUG")
                        if _verify_restored_venv(venv_path=venv_dst, deps=deps_clean, run_cmd=run_cmd):
                            return sandbox
                        dbg("Cached venv appears corrupted - rebuilding", "WARNING")
                        try:
                            shutil.rmtree(venv_dst)
                        except Exception:
                            pass
                        try:
                            if self._dep_cache:
                                self._dep_cache.invalidate(deps_clean)
                        except Exception:
                            pass
                    except Exception:
                        try:
                            stop.set()
                        except Exception:
                            pass

                dbg(f"Creating venv (.venv) in sandbox", "INFO")
                try:
                    stop = Event()
                    thr = Thread(
                        target=_heartbeat,
                        kwargs={
                            "stop": stop,
                            "on_log": on_log,
                            "message": f"[deploy] Creating venv (.venv) ({len(deps_clean)} deps)",
                            "interval_s": float(_beat_every_s()),
                        },
                        daemon=True,
                    )
                    thr.start()
                    ensure_venv(sandbox, verbose=False)
                    stop.set()
                    dbg(f"Venv status: {_path_debug(sandbox.path / '.venv')}", "DEBUG")
                except Exception as e:
                    try:
                        stop.set()
                    except Exception:
                        pass
                    dbg(f"ensure_venv failed: {e}", "ERROR")
                    raise
                dbg("Installing dependencies via pip", "INFO")
                try:
                    pip_stop = Event()
                    thr = Thread(
                        target=_heartbeat,
                        kwargs={
                            "stop": pip_stop,
                            "on_log": on_log,
                            "message": f"[deploy] Installing dependencies via pip ({len(deps_clean)} deps)",
                            "interval_s": float(_beat_every_s()),
                        },
                        daemon=True,
                    )
                    thr.start()
                    install_env = os.environ.copy()
                    for k, v in (env or {}).items():
                        if k is None or v is None:
                            continue
                        install_env[str(k)] = str(v)

                    pip_path = sandbox.venv_bin / "pip"
                    requirements_path = sandbox.path / "requirements.txt"

                    try:
                        proc = subprocess.Popen(
                            [
                                str(pip_path),
                                "install",
                                "--disable-pip-version-check",
                                "--progress-bar",
                                "off",
                                "-r",
                                str(requirements_path),
                            ],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            text=True,
                            bufsize=1,
                            env=install_env,
                        )
                        if proc.stdout:
                            for line in proc.stdout:
                                s = (line or "").rstrip("\n")
                                if not s:
                                    continue
                                if on_log:
                                    _call_on_log(on_log, s, "INFO")
                        rc = proc.wait()
                        if rc != 0:
                            raise subprocess.CalledProcessError(rc, proc.args)
                    except subprocess.CalledProcessError as e:
                        dbg(f"pip install failed: {e}", "ERROR")
                        raise
                    finally:
                        try:
                            pip_stop.set()
                        except Exception:
                            pass
                    dbg("Dependencies installed", "INFO")
                    try:
                        self._dep_cache.save_existing_venv(deps_clean, sandbox.path / ".venv", on_progress=on_log)
                    except Exception:
                        pass
                except Exception as e:
                    try:
                        pip_stop.set()
                    except Exception:
                        pass
                    dbg(f"install_deps failed: {e}", "ERROR")
                    raise
        else:
            dbg("No dependencies block found", "DEBUG")

        return sandbox

    def start_service(
        self,
        service: ServiceConfig,
        readme_path: Path,
        env: dict[str, str],
        verbose: bool = True,
        restart_if_running: bool = False,
        on_log: Optional[Callable[[str], None]] = None,
        user_id: Optional[str] = None,
    ) -> ServiceProcess:
        """Start a service in its sandbox.
        
        Args:
            service: Service configuration
            readme_path: Path to README.md with markpact blocks
            env: Environment variables to pass to the service
            verbose: Print status messages
            restart_if_running: If True, stop and restart if already running
            on_log: Callback for detailed logging
            user_id: Optional SaaS user ID for process isolation
        """
        def log(msg: str, level: str = "INFO"):
            logger.log(getattr(logging, level), f"[{service.name}] {msg}")
            if on_log and _should_emit_to_ui(level):
                _call_on_log(on_log, msg, level)
            if verbose and _should_emit_to_ui(level):
                print(msg)
        
        log(f"Starting service: {service.name}", "INFO")
        log(f"Port: {service.port}, README: {readme_path}", "DEBUG")
        log(f"Runner UID/EUID/GID: uid={os.getuid()} euid={os.geteuid()} gid={os.getgid()}", "DEBUG")
        log(f"Sandbox root: {_path_debug(self.sandbox_root)}", "DEBUG")
        log(f"README: {_path_debug(readme_path)}", "DEBUG")
        
        if service.name in self._processes:
            existing = self._processes[service.name]
            if existing.is_running:
                if restart_if_running:
                    log(f"Restarting {service.name}...", "INFO")
                    self.stop_service(service.name)
                    self.clean_sandbox(service.name)
                else:
                    log(f"Service {service.name} already running", "ERROR")
                    raise RuntimeError(f"Service {service.name} is already running")

        # Create sandbox with dependency installation
        log("Creating sandbox and installing dependencies...", "INFO")
        try:
            sandbox = self.create_sandbox(
                service,
                readme_path,
                install_dependencies=True,
                on_log=log,
                env=env,
            )
            log(f"Sandbox created at: {sandbox.path}", "INFO")
        except Exception as e:
            log(f"Failed to create sandbox: {e}", "ERROR")
            logger.exception(f"Sandbox creation failed for {service.name}")
            raise

        readme_content = readme_path.read_text()
        blocks = parse_blocks(readme_content)

        run_command = None
        for block in blocks:
            if block.kind == "run":
                run_command = block.body.strip()
                break

        if not run_command:
            log(f"No run command found in README", "ERROR")
            raise ValueError(f"No run command found in {readme_path}")

        log(f"Run command: {run_command}", "DEBUG")

        full_env = os.environ.copy()
        full_env.update(env)
        
        # Ensure PORT is always set in environment
        full_env["PORT"] = str(service.port)
        full_env["MARKPACT_PORT"] = str(service.port)

        dotenv_env = dict(env or {})
        dotenv_env["PORT"] = str(service.port)
        dotenv_env["MARKPACT_PORT"] = str(service.port)
        _write_dotenv_file(sandbox.path, dotenv_env)

        if sandbox.has_venv():
            venv_bin = str(sandbox.venv_bin)
            full_env["PATH"] = f"{venv_bin}:{full_env.get('PATH', '')}"
            full_env["VIRTUAL_ENV"] = str(sandbox.path / ".venv")
            log(f"Using venv: {sandbox.path / '.venv'}", "DEBUG")
        else:
            log("WARNING: No venv found, using system Python", "WARNING")

        # Expand $PORT in command
        expanded_cmd = run_command.replace("$PORT", str(service.port))
        expanded_cmd = expanded_cmd.replace("${PORT}", str(service.port))
        expanded_cmd = expanded_cmd.replace("${MARKPACT_PORT}", str(service.port))
        expanded_cmd = expanded_cmd.replace("$MARKPACT_PORT", str(service.port))
        
        # Replace hardcoded ports in run command with the requested port
        # This handles cases where LLM generates hardcoded ports like --port 8000
        import re
        port_patterns = [
            (r'--port[=\s]+(\d+)', f'--port {service.port}'),  # --port 8000 or --port=8000
            (r'-p[=\s]+(\d+)', f'-p {service.port}'),          # -p 8000 or -p=8000
            (r':(\d{4,5})(?=\s|$|")', f':{service.port}'),     # :8000 at end of string
        ]
        
        original_cmd = expanded_cmd
        for pattern, replacement in port_patterns:
            match = re.search(pattern, expanded_cmd)
            if match:
                old_port = match.group(1) if match.groups() else None
                if old_port and old_port != str(service.port):
                    log(f"Replacing hardcoded port {old_port} with {service.port}", "INFO")
                    expanded_cmd = re.sub(pattern, replacement, expanded_cmd)
        
        if expanded_cmd != original_cmd:
            log(f"Port-corrected command: {expanded_cmd}", "INFO")
        else:
            log(f"Expanded command: {expanded_cmd}", "DEBUG")
        
        # Remove --reload flag from uvicorn commands in sandbox environments
        # --reload uses multiprocessing which can crash in Docker containers
        if "--reload" in expanded_cmd and "uvicorn" in expanded_cmd:
            expanded_cmd = re.sub(r'\s*--reload\s*', ' ', expanded_cmd)
            log(f"Removed --reload flag (not compatible with sandbox): {expanded_cmd}", "INFO")

        if sandbox.has_venv():
            venv_python = sandbox.venv_bin / "python"
            venv_python_q = shlex.quote(str(venv_python))

            # Prefer venv python for common Python entrypoints. This is more robust than
            # relying on PATH when running under user isolation.
            rewritten = expanded_cmd
            rewritten = re.sub(r"^\s*uvicorn(\s+)", rf"{venv_python_q} -m uvicorn\1", rewritten, count=1)
            rewritten = re.sub(r"^\s*gunicorn(\s+)", rf"{venv_python_q} -m gunicorn\1", rewritten, count=1)
            rewritten = re.sub(r"^\s*python3?(\s+)", rf"{venv_python_q}\1", rewritten, count=1)

            if rewritten != expanded_cmd:
                expanded_cmd = rewritten
                log(f"Rewriting run command to use venv python: {expanded_cmd}", "DEBUG")

        log(f"Starting process...", "INFO")

        # Use user isolation if user_id provided
        preexec = os.setsid
        if user_id:
            try:
                from .user_isolation import get_isolation_manager
                isolation = get_isolation_manager()
                try:
                    can_isolate, reason = isolation.can_isolate()
                    log(f"Isolation capability: can_isolate={can_isolate} reason={reason}", "DEBUG")
                except Exception as e:
                    log(f"Isolation capability check failed: {e}", "WARNING")
                user = isolation.get_or_create_user(user_id)
                log(f"🔒 Running as isolated user: {user.linux_username} (uid={user.linux_uid})", "INFO")
                
                # Update env with user-specific settings
                full_env["HOME"] = str(user.home_dir)
                full_env["USER"] = user.linux_username
                full_env["LOGNAME"] = user.linux_username
                
                # Create preexec function for user switching
                def preexec():
                    os.setsid()
                    if os.geteuid() == 0:
                        os.setgid(user.linux_gid)
                        os.setuid(user.linux_uid)
            except Exception as e:
                log(f"⚠️ User isolation not available: {e} - using sandbox uid", "WARNING")
                if os.geteuid() == 0:
                    uid, gid = _sandbox_fallback_ids()
                    log(f"Sandbox uid fallback: uid={uid} gid={gid}", "DEBUG")
                    try:
                        dotenv_path = sandbox.path / ".env"
                        if dotenv_path.exists():
                            os.chown(dotenv_path, uid, gid)
                    except Exception:
                        pass
                    _chown_sandbox_tree(sandbox.path, uid, gid)

                    def preexec():
                        os.setsid()
                        os.setgid(gid)
                        os.setuid(uid)

        # Always capture stderr for debugging
        # nosec B602: shell=True required - we execute user-defined run commands
        # Input is validated via markpact parsing and sandbox isolation
        process = subprocess.Popen(
            expanded_cmd,
            shell=True,  # nosec B602
            cwd=str(sandbox.path),
            env=full_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=preexec,
        )

        log(f"Process started with PID: {process.pid}", "INFO")

        # Wait briefly for process to start (but don't block too long)
        time.sleep(0.2)
        
        # Check if process died immediately
        poll_result = process.poll()
        if poll_result is not None:
            # Process already died - capture all output
            exit_code = poll_result
            stderr = ""
            stdout = ""
            
            try:
                # Read all output with timeout
                stdout_bytes, stderr_bytes = process.communicate(timeout=2)
                stderr = stderr_bytes.decode('utf-8', errors='replace') if stderr_bytes else ""
                stdout = stdout_bytes.decode('utf-8', errors='replace') if stdout_bytes else ""
            except Exception as e:
                log(f"Could not read process output: {e}", "WARNING")
                if process.stderr:
                    try:
                        stderr = process.stderr.read().decode('utf-8', errors='replace')
                    except:
                        pass
            
            # Interpret exit code
            if exit_code < 0:
                signal_name = {
                    -9: "SIGKILL",
                    -15: "SIGTERM", 
                    -11: "SIGSEGV",
                    -6: "SIGABRT",
                }.get(exit_code, f"signal {-exit_code}")
                log(f"Process killed by {signal_name} (exit code: {exit_code})", "ERROR")
            else:
                log(f"Process exited with code: {exit_code}", "ERROR")
            
            if stderr:
                log(f"STDERR:\n{stderr[:2000]}", "ERROR")
            if stdout:
                log(f"STDOUT:\n{stdout[:1000]}", "DEBUG")
            
            # Write to error log file
            error_log = LOG_DIR / f"{service.name}_error.log"
            with open(error_log, "w") as f:
                f.write(f"Exit code: {exit_code}\n")
                f.write(f"Command: {expanded_cmd}\n")
                f.write(f"CWD: {sandbox.path}\n")
                f.write(f"Venv: {sandbox.path / '.venv'}\n")
                f.write(f"\n--- STDERR ---\n{stderr}\n")
                f.write(f"\n--- STDOUT ---\n{stdout}\n")
                # List files for debugging
                try:
                    files = list(sandbox.path.glob("*"))
                    f.write(f"\n--- FILES ---\n{[str(f) for f in files]}\n")
                except:
                    pass
            log(f"Error log written to: {error_log}", "DEBUG")

        svc_process = ServiceProcess(
            name=service.name,
            pid=process.pid,
            port=service.port,
            sandbox_path=sandbox.path,
            process=process,
        )

        self._processes[service.name] = svc_process
        
        # Log sandbox contents for debugging
        try:
            files = list(sandbox.path.glob("*"))
            log(f"Sandbox files: {[f.name for f in files]}", "DEBUG")
        except:
            pass
        
        return svc_process

    def stop_service(self, service_name: str, timeout: int = 10) -> bool:
        """Stop a running service."""
        if service_name not in self._processes:
            logger.debug(f"Service {service_name} not in tracked processes")
            return False

        svc = self._processes[service_name]
        old_pid = svc.pid

        if not svc.is_running:
            logger.debug(f"Service {service_name} (PID {old_pid}) already stopped")
            del self._processes[service_name]
            return True

        logger.info(f"Stopping service {service_name} (PID {old_pid})")
        
        try:
            pgid = os.getpgid(old_pid)
            os.killpg(pgid, signal.SIGTERM)
            logger.debug(f"Sent SIGTERM to process group {pgid}")
        except ProcessLookupError:
            logger.debug(f"Process {old_pid} already gone")
            del self._processes[service_name]
            return True
        except OSError as e:
            logger.warning(f"Error getting pgid for {old_pid}: {e}")
            # Try killing just the process
            try:
                os.kill(old_pid, signal.SIGTERM)
            except ProcessLookupError:
                pass

        deadline = time.time() + timeout
        while time.time() < deadline:
            if not svc.is_running:
                break
            time.sleep(0.1)

        if svc.is_running:
            logger.warning(f"Service {service_name} didn't stop gracefully, sending SIGKILL")
            try:
                os.killpg(os.getpgid(old_pid), signal.SIGKILL)
            except (ProcessLookupError, OSError):
                pass

        del self._processes[service_name]
        
        # Wait for OS to clean up the process
        time.sleep(0.3)
        logger.info(f"Service {service_name} stopped")
        return True

    def stop_all(self, timeout: int = 10) -> None:
        """Stop all running services."""
        for name in list(self._processes.keys()):
            self.stop_service(name, timeout)

    def get_status(self, service_name: str) -> Optional[dict]:
        """Get status of a service."""
        if service_name not in self._processes:
            return None

        svc = self._processes[service_name]
        return {
            "name": svc.name,
            "pid": svc.pid,
            "port": svc.port,
            "running": svc.is_running,
            "uptime": time.time() - svc.started_at,
            "sandbox": str(svc.sandbox_path),
        }

    def get_all_status(self) -> list[dict]:
        """Get status of all services."""
        return [
            self.get_status(name)
            for name in self._processes
            if self.get_status(name)
        ]

    def clean_sandbox(self, service_name: str) -> None:
        """Remove sandbox directory for a service."""
        sandbox_path = self.get_sandbox_path(service_name)
        if sandbox_path.exists():
            shutil.rmtree(sandbox_path)

    def clean_all(self) -> None:
        """Remove all sandbox directories."""
        if self.sandbox_root.exists():
            shutil.rmtree(self.sandbox_root)
        self.sandbox_root.mkdir(parents=True)

    def create_sandboxes_parallel(
        self,
        services: list[tuple[ServiceConfig, Path]],
        max_workers: int = 4,
        on_complete: Optional[Callable[[str, bool, float], None]] = None,
    ) -> dict[str, Sandbox]:
        """
        Create sandboxes for multiple services in parallel.

        Args:
            services: List of (ServiceConfig, readme_path) tuples
            max_workers: Maximum parallel workers
            on_complete: Callback(name, success, duration)

        Returns:
            Dict of {service_name: Sandbox}
        """
        results: dict[str, Sandbox] = {}
        errors: dict[str, str] = {}
        lock = Lock()

        def create_one(service: ServiceConfig, readme_path: Path) -> tuple[str, Sandbox]:
            sandbox = self.create_sandbox(service, readme_path)
            return service.name, sandbox

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {}
            start_times = {}

            for service, readme_path in services:
                start_times[service.name] = time.time()
                future = executor.submit(create_one, service, readme_path)
                futures[future] = service.name

            for future in as_completed(futures):
                name = futures[future]
                duration = time.time() - start_times[name]

                try:
                    _, sandbox = future.result()
                    with lock:
                        results[name] = sandbox
                    if on_complete:
                        on_complete(name, True, duration)
                except Exception as e:
                    with lock:
                        errors[name] = str(e)
                    if on_complete:
                        on_complete(name, False, duration)

        if errors:
            error_msg = "; ".join(f"{k}: {v}" for k, v in errors.items())
            raise RuntimeError(f"Failed to create sandboxes: {error_msg}")

        return results

    def start_services_parallel(
        self,
        services: list[tuple[ServiceConfig, Path, dict[str, str]]],
        max_workers: int = 4,
        on_complete: Optional[Callable[[str, bool, float], None]] = None,
    ) -> dict[str, ServiceProcess]:
        """
        Start multiple services in parallel.

        Note: Should only be used for services with no inter-dependencies.
        For dependent services, use the orchestrator's wave-based approach.

        Args:
            services: List of (ServiceConfig, readme_path, env) tuples
            max_workers: Maximum parallel workers
            on_complete: Callback(name, success, duration)

        Returns:
            Dict of {service_name: ServiceProcess}
        """
        results: dict[str, ServiceProcess] = {}
        errors: dict[str, str] = {}
        lock = Lock()

        def start_one(
            service: ServiceConfig,
            readme_path: Path,
            env: dict[str, str]
        ) -> tuple[str, ServiceProcess]:
            proc = self.start_service(service, readme_path, env, verbose=False)
            return service.name, proc

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {}
            start_times = {}

            for service, readme_path, env in services:
                start_times[service.name] = time.time()
                future = executor.submit(start_one, service, readme_path, env)
                futures[future] = service.name

            for future in as_completed(futures):
                name = futures[future]
                duration = time.time() - start_times[name]

                try:
                    _, proc = future.result()
                    with lock:
                        results[name] = proc
                    if on_complete:
                        on_complete(name, True, duration)
                except Exception as e:
                    with lock:
                        errors[name] = str(e)
                    if on_complete:
                        on_complete(name, False, duration)

        return results, errors



"""Docker Compose / Podman Compose generator for multi-service deployment."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import yaml

from ..config import EcosystemConfig, ServiceConfig
from .base import DeploymentConfig, DeploymentMode


@dataclass
class ComposeService:
    """Represents a service in docker-compose.yaml."""
    name: str
    build_context: str
    dockerfile: str
    image: str
    ports: list[str]
    environment: dict[str, str]
    depends_on: list[str]
    health_check: Optional[dict] = None
    deploy: Optional[dict] = None
    networks: list[str] = None
    volumes: list[str] = None


class ComposeGenerator:
    """
    Generate Docker Compose / Podman Compose files for pactown ecosystems.

    Supports:
    - Docker Compose v3.8+
    - Podman Compose
    - Docker Swarm mode
    """

    def __init__(
        self,
        ecosystem: EcosystemConfig,
        deploy_config: DeploymentConfig,
        base_path: Path,
    ):
        self.ecosystem = ecosystem
        self.deploy_config = deploy_config
        self.base_path = Path(base_path)

    def generate(
        self,
        output_path: Optional[Path] = None,
        include_registry: bool = False,
    ) -> dict:
        """
        Generate docker-compose.yaml content.

        Args:
            output_path: Optional path to write the file
            include_registry: Include pactown registry service

        Returns:
            Compose file as dict
        """
        compose = {
            "version": "3.8",
            "name": self.ecosystem.name,
            "services": {},
            "networks": {
                self.deploy_config.network_name: {
                    "driver": "bridge",
                },
            },
        }

        # Add volumes section if needed
        volumes = {}

        for name, service in self.ecosystem.services.items():
            compose_service = self._create_service(name, service)
            compose["services"][name] = compose_service

            # Check for volume mounts
            if self.deploy_config.volumes_path:
                volume_name = f"{name}-data"
                volumes[volume_name] = {"driver": "local"}

        if volumes:
            compose["volumes"] = volumes

        # Add registry service if requested
        if include_registry:
            compose["services"]["registry"] = self._create_registry_service()

        # Write to file if path provided
        if output_path:
            output_path = Path(output_path)
            with open(output_path, "w") as f:
                yaml.dump(compose, f, default_flow_style=False, sort_keys=False)

        return compose

    def _create_service(self, name: str, service: ServiceConfig) -> dict:
        """Create compose service definition."""
        self.base_path / service.readme
        sandbox_path = self.base_path / self.ecosystem.sandbox_root / name

        svc = {
            "build": {
                "context": str(sandbox_path),
                "dockerfile": "Dockerfile",
            },
            "image": f"{self.deploy_config.image_prefix}/{name}:latest",
            "container_name": f"{self.ecosystem.name}-{name}",
            "restart": "unless-stopped",
            "networks": [self.deploy_config.network_name],
        }

        # Ports
        if service.port and self.deploy_config.expose_ports:
            svc["ports"] = [f"{service.port}:{service.port}"]

        # Environment
        env = {"SERVICE_NAME": name}
        if service.port:
            env["MARKPACT_PORT"] = str(service.port)

        # Add dependency URLs
        for dep in service.depends_on:
            dep_service = self.ecosystem.services.get(dep.name)
            if dep_service:
                env_key = dep.name.upper().replace("-", "_")
                # Use container DNS name
                env[f"{env_key}_URL"] = f"http://{dep.name}:{dep_service.port}"

        env.update(service.env)
        svc["environment"] = env

        # Dependencies
        if service.depends_on:
            svc["depends_on"] = {}
            for dep in service.depends_on:
                if dep.name in self.ecosystem.services:
                    dep_svc = self.ecosystem.services[dep.name]
                    condition = "service_healthy" if dep_svc.health_check else "service_started"
                    svc["depends_on"][dep.name] = {"condition": condition}

        # Health check
        if service.health_check:
            svc["healthcheck"] = {
                "test": ["CMD", "curl", "-f", f"http://localhost:{service.port}{service.health_check}"],
                "interval": self.deploy_config.health_check_interval,
                "timeout": self.deploy_config.health_check_timeout,
                "retries": self.deploy_config.health_check_retries,
                "start_period": "10s",
            }

        # Production settings
        if self.deploy_config.mode == DeploymentMode.PRODUCTION:
            svc["deploy"] = {
                "resources": {
                    "limits": {
                        "memory": self.deploy_config.memory_limit,
                        "cpus": self.deploy_config.cpu_limit,
                    },
                    "reservations": {
                        "memory": "128M",
                        "cpus": "0.1",
                    },
                },
                "restart_policy": {
                    "condition": "on-failure",
                    "delay": "5s",
                    "max_attempts": 3,
                    "window": "120s",
                },
            }

            # Security options
            svc["security_opt"] = ["no-new-privileges:true"]

            if self.deploy_config.read_only_fs:
                svc["read_only"] = True
                svc["tmpfs"] = ["/tmp"]

            svc["cap_drop"] = self.deploy_config.drop_capabilities

            if self.deploy_config.add_capabilities:
                svc["cap_add"] = self.deploy_config.add_capabilities

        # Labels
        svc["labels"] = {
            "pactown.ecosystem": self.ecosystem.name,
            "pactown.service": name,
            **self.deploy_config.labels,
        }

        return svc

    def _create_registry_service(self) -> dict:
        """Create pactown registry service."""
        return {
            "image": "pactown/registry:latest",
            "container_name": f"{self.ecosystem.name}-registry",
            "restart": "unless-stopped",
            "ports": ["8800:8800"],
            "networks": [self.deploy_config.network_name],
            "volumes": ["registry-data:/data"],
            "environment": {
                "REGISTRY_PORT": "8800",
                "REGISTRY_DATA_DIR": "/data",
            },
            "healthcheck": {
                "test": ["CMD", "curl", "-f", "http://localhost:8800/health"],
                "interval": "30s",
                "timeout": "10s",
                "retries": 3,
            },
        }

    def generate_override(
        self,
        output_path: Optional[Path] = None,
        dev_mode: bool = True,
    ) -> dict:
        """
        Generate docker-compose.override.yaml for development.

        This file is automatically merged with docker-compose.yaml
        and provides development-specific settings.
        """
        override = {
            "version": "3.8",
            "services": {},
        }

        for name, service in self.ecosystem.services.items():
            sandbox_path = self.base_path / self.ecosystem.sandbox_root / name

            svc = {}

            if dev_mode:
                # Mount source code for hot reload
                svc["volumes"] = [
                    f"{sandbox_path}:/app:z",
                ]

                # Enable debug mode
                svc["environment"] = {
                    "DEBUG": "true",
                    "LOG_LEVEL": "debug",
                }

                # Remove resource limits for development
                svc["deploy"] = None

            if svc:
                override["services"][name] = svc

        if output_path:
            output_path = Path(output_path)
            with open(output_path, "w") as f:
                yaml.dump(override, f, default_flow_style=False, sort_keys=False)

        return override

    def generate_production(
        self,
        output_path: Optional[Path] = None,
        replicas: int = 2,
    ) -> dict:
        """
        Generate docker-compose.prod.yaml for production/swarm deployment.
        """
        compose = self.generate()

        for name in compose["services"]:
            svc = compose["services"][name]

            # Add swarm-specific deploy config
            svc["deploy"] = {
                "mode": "replicated",
                "replicas": replicas,
                "update_config": {
                    "parallelism": 1,
                    "delay": "10s",
                    "failure_action": "rollback",
                    "order": "start-first",
                },
                "rollback_config": {
                    "parallelism": 1,
                    "delay": "10s",
                },
                "resources": {
                    "limits": {
                        "memory": self.deploy_config.memory_limit,
                        "cpus": self.deploy_config.cpu_limit,
                    },
                },
            }

        if output_path:
            output_path = Path(output_path)
            with open(output_path, "w") as f:
                yaml.dump(compose, f, default_flow_style=False, sort_keys=False)

        return compose


def generate_compose_from_config(
    config_path: Path,
    output_dir: Optional[Path] = None,
    production: bool = False,
) -> dict:
    """
    Convenience function to generate compose files from pactown config.

    Args:
        config_path: Path to saas.pactown.yaml
        output_dir: Directory to write compose files
        production: Generate production configuration

    Returns:
        Generated compose dict
    """
    from ..config import load_config

    config_path = Path(config_path)
    ecosystem = load_config(config_path)

    deploy_config = (
        DeploymentConfig.for_production() if production
        else DeploymentConfig.for_development()
    )

    generator = ComposeGenerator(
        ecosystem=ecosystem,
        deploy_config=deploy_config,
        base_path=config_path.parent,
    )

    output_dir = output_dir or config_path.parent

    # Generate main compose file
    compose = generator.generate(
        output_path=output_dir / "docker-compose.yaml"
    )

    # Generate override for development
    if not production:
        generator.generate_override(
            output_path=output_dir / "docker-compose.override.yaml"
        )
    else:
        generator.generate_production(
            output_path=output_dir / "docker-compose.prod.yaml"
        )

    return compose



from __future__ import annotations

import re
from typing import Literal, Optional

from pydantic import BaseModel, Field, field_validator

SubdomainSeparator = Literal["-", "."]


def coerce_subdomain_separator(value: Optional[str]) -> SubdomainSeparator:
    return "." if value == "." else "-"


def normalize_host(value: str) -> str:
    v = (value or "").strip().lower()
    if v.startswith("http://"):
        v = v[len("http://") :]
    elif v.startswith("https://"):
        v = v[len("https://") :]
    v = v.split("/", 1)[0]
    v = v.split(":", 1)[0]
    v = v.strip(".")
    return v


def normalize_domain(value: str) -> str:
    v = normalize_host(value)
    if v.startswith("www."):
        v = v[len("www.") :]
    return v


def is_local_domain(domain: str) -> bool:
    d = normalize_domain(domain)
    return d in {"localhost", "127.0.0.1", "0.0.0.0"}


def build_origin(*, scheme: Literal["http", "https"], host: str, port: Optional[int] = None) -> str:
    h = normalize_host(host)
    if port is None:
        return f"{scheme}://{h}"
    return f"{scheme}://{h}:{int(port)}"


def web_base_url(domain: str, web_host_port: int) -> str:
    if is_local_domain(domain):
        return build_origin(scheme="http", host="localhost", port=int(web_host_port))
    return build_origin(scheme="https", host=normalize_domain(domain))


def api_base_url(domain: str, api_host_port: int) -> str:
    if is_local_domain(domain):
        return build_origin(scheme="http", host="localhost", port=int(api_host_port))
    return build_origin(scheme="https", host=f"api.{normalize_domain(domain)}")


def to_dns_label(value: str, *, max_len: int = 63, fallback: str = "x") -> str:
    v = (value or "").lower().strip()
    v = re.sub(r"[^a-z0-9]+", "-", v)
    v = re.sub(r"-+", "-", v).strip("-")
    if not v:
        v = fallback
    return v[:max_len]


class DomainConfig(BaseModel):
    domain: str = Field(default="localhost")
    subdomain_separator: SubdomainSeparator = Field(default="-")

    @field_validator("domain")
    @classmethod
    def _normalize_domain(cls, v: str) -> str:
        out = normalize_domain(v)
        if not out:
            return "localhost"
        return out

    @field_validator("subdomain_separator")
    @classmethod
    def _normalize_separator(cls, v: str) -> SubdomainSeparator:
        return coerce_subdomain_separator(v)


class ProjectHostParts(BaseModel):
    project_id: int
    username: str


_PROJECT_SUBDOMAIN_RE = re.compile(r"^(?P<project_id>\d+)(?:-|\.)(?P<username>[a-z0-9-]+)$")


def parse_project_subdomain(subdomain: str) -> Optional[ProjectHostParts]:
    s = (subdomain or "").strip().lower()
    m = _PROJECT_SUBDOMAIN_RE.match(s)
    if not m:
        return None
    try:
        project_id = int(m.group("project_id"))
    except Exception:
        return None
    username = m.group("username")
    if not username:
        return None
    return ProjectHostParts(project_id=project_id, username=username)


def build_project_subdomain(project_id: int, username: str, *, separator: SubdomainSeparator) -> str:
    pid = int(project_id)
    sep = coerce_subdomain_separator(separator)

    prefix = f"{pid}{sep}"
    max_username_len = max(1, 63 - len(prefix))
    uname = to_dns_label(username, max_len=max_username_len, fallback="user")

    return f"{pid}{sep}{uname}"


def build_project_host(project_id: int, username: str, *, domain: str, separator: SubdomainSeparator) -> str:
    base_domain = normalize_domain(domain)
    sub = build_project_subdomain(project_id, username, separator=separator)
    return f"{sub}.{base_domain}"


def parse_project_host(host: str, *, domain: str) -> Optional[ProjectHostParts]:
    h = normalize_host(host)
    base_domain = normalize_domain(domain)
    suffix = f".{base_domain}" if base_domain else ""

    if not suffix or not h.endswith(suffix):
        return None

    subdomain = h[: -len(suffix)]
    return parse_project_subdomain(subdomain)


def build_service_subdomain(service_name: str, username: str, *, separator: SubdomainSeparator) -> str:
    sep = coerce_subdomain_separator(separator)
    service_label = to_dns_label(service_name, max_len=20, fallback="app")
    tenant_label = to_dns_label(username, max_len=30, fallback="user")

    if sep == ".":
        return f"{service_label}.{tenant_label}".strip(".")

    label = f"{service_label}-{tenant_label}".strip("-")
    return label[:63].strip("-")
